import os
import time
import numpy as np
import pandas as pd
import xarray as xr
import glob as glob
from sys import argv

# from netCDF4 import Dataset
import math
import matplotlib.pyplot as plt
from datetime import date
import time
import multiprocessing
import warnings

# warnings.filterwarnings("ignore")  # TODO don't use normally!!!


def dealiaze(data):
    n_cores = 16  # multiprocessing.cpu_count()
    # if n_cores > 1:
    # 	n_cores = n_cores - 1 # we have the main function running on one core and our institute does not allow to use all cores
    breakpoint()
    pool = multiprocessing.Pool(n_cores)
    Nyquvist = data.nyquist_velocity.values  # [6.350626, 4.981658, 2.61496353]
    RangeEnd = data.maximum_altitude.values  # [715.4785, 5902.698, 17994.29]
    doppLen = data.doppLen.values #TODO: check if this is range_vectors or len(range_vectors)
    vel = data.velocity_vectors.values 
    dv = np.diff(vel)[0]
    newVel = np.arange(-Nyquvist[0] * 2, Nyquvist[0] * 2 + dv, dv, dtype=np.float32)


    data["minVelH"] = xr.DataArray(
        dims=("range", "Time"), coords={"range": data.range, "Time": data.Time}
    )
    data["maxVelH"] = data["minVelH"].copy()
    data["minVelV"] = data["minVelH"].copy()
    data["maxVelV"] = data["minVelH"].copy()
    data["minVelZDR"] = data["minVelH"].copy()
    data["maxVelZDR"] = data["minVelH"].copy()
    data["HSpec_rot"] = xr.DataArray(
        dims=("Time", "range", "vel_rot"),
        coords={"vel_rot": newVel, "Time": data.Time, "range": data.range},
    )
    data["VSpec_rot"] = data["HSpec_rot"].copy()
    data["sZDR_rot"] = data["HSpec_rot"].copy()
    # - testing dealiasing:
    for tind, t in enumerate(data.Time):
        print(t.values)
        # print(data)
        dataRangeValues = data.range.where(
            np.isfinite(10 * np.log10(data.sel(Time=t).ZH)), drop=True
        )
        args = [(r, data.sel(Time=t)) for r in dataRangeValues]
        for (
            r,
            HSpec,
            VSpec,
            sZDR,
            sSNR_H,
            sSNR_V,
            maxVelH,
            minVelH,
            maxVelV,
            minVelV,
            maxVelZDR,
            minVelZDR,
            HSpec_rot,
            VSpec_rot,
            sZDR_rot,
        ) in pool.starmap(dealiazeOneHeight, args):
            data["HSpec"].loc[t, r, :] = HSpec
            data["VSpec"].loc[t, r, :] = VSpec
            data["sZDR"].loc[t, r, :] = sZDR
            data["sSNR_H"].loc[t, r, :] = sSNR_H
            data["sSNR_V"].loc[t, r, :] = sSNR_V
            data["maxVelH"].loc[r, t] = maxVelH
            data["minVelH"].loc[r, t] = minVelH
            data["maxVelV"].loc[r, t] = maxVelV
            data["minVelV"].loc[r, t] = minVelV
            data["maxVelZDR"].loc[r, t] = maxVelZDR
            data["minVelZDR"].loc[r, t] = minVelZDR
            data["HSpec_rot"].loc[t, r, :] = HSpec_rot
            data["VSpec_rot"].loc[t, r, :] = VSpec_rot
            data["sZDR_rot"].loc[t, r, :] = sZDR_rot

    return data

def dealiazeOneHeight(r,data):
	Nyquvist = data.nyquist_velocity.values#[6.350626, 4.981658, 2.61496353]
	RangeEnd = data.maximum_altitude.values#[715.4785, 5902.698, 17994.29]
	doppLen = data.doppLen.values
	vel = data.velocity_vectors.values
	dv = np.diff(vel)[0]
	newVel = np.arange(-Nyquvist[0]*2,Nyquvist[0]*2+dv,dv,dtype=np.float32)
	#print(r)
	datasel = data.sel(range=r)
	#datasel = datasel[['sZDR','HSpec','VSpec','sSNR_H','sSNR_V']]
	sZDR = datasel.sZDR
	HSpec = datasel.HSpec
	VSpec = datasel.VSpec
	sSNR_H = datasel.sSNR_H
	sSNR_V = datasel.sSNR_V
	if r > RangeEnd[1]: # get correct nyquvist vel (I already regridded everything so I need to do it like tha
		maxVel = Nyquvist[2]
		ra = RangeEnd[2]
	elif r <= RangeEnd[0]:
		maxVel = Nyquvist[0]
		ra = RangeEnd[0]
	else:
		maxVel = Nyquvist[1]
		ra = RangeEnd[1]
	
	if (~np.isnan(sZDR.sel(Vel=maxVel,method='nearest')) or ~np.isnan(sZDR.sel(Vel=-maxVel,method='nearest')) or ~np.isnan(sZDR.sel(Vel=-maxVel+dv,method='nearest')) 
		or ~np.isnan(sZDR.sel(Vel=maxVel-dv,method='nearest')) or ~np.isnan(sZDR.sel(Vel=-maxVel+2*dv,method='nearest')) or ~np.isnan(sZDR.sel(Vel=maxVel-2*dv,method='nearest')) 
		or ~np.isnan(sZDR.sel(Vel=-maxVel+3*dv,method='nearest')) or ~np.isnan(sZDR.sel(Vel=maxVel-3*dv,method='nearest'))): 
	# if we have non-noise values at +- Nyquvist range or upto 3 vel bins to the right or left
		datadrop = datasel.where(data.Vel < maxVel,drop=True) #- since we regridded, I need to drop everything that is larger or smaller than +- Ny to actually kit the spectra together (otherwise we would have a lot of nan in between..)
		datadrop = datadrop.where(data.Vel > -maxVel,drop=True)
		q = len(datadrop.Vel.values)-1;
		k = 0;
		
		if np.isnan(datadrop.HSpec.isel(Vel=q)):
			datadrop.HSpec.values[q] = -99999
		for i in range(q):
			if np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+1])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+2])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+3])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+4])):
				datadrop.HSpec.values[i] = -99999
			elif np.isnan(datadrop.HSpec.isel(Vel=[i])) and  ~np.isnan(datadrop.HSpec.isel(Vel=[i+5])):
				datadrop.HSpec.values[i] = -99999
		if datadrop.HSpec.isnull().any(dim='Vel'):
			while ~np.isnan(datadrop.HSpec.isel(Vel=q)): #-- find out how much we need to shift
				k = k + 1;
				q = q - 1;
		# now roll the coordinates:
		datadrop = datadrop.where(datadrop.HSpec!=-99999,np.nan)
		#print(maxVel)
		dataRoll = datadrop.roll(Vel=k+1,roll_coords=False)  # shift everything, the direction doesnt matter
		
		dataRollRegridded = dataRoll.reindex({'Vel':data.Vel.values}, method = 'nearest', tolerance = 0.05) # now I need to go back to original velocity
		HSpec = dataRollRegridded.HSpec
		VSpec = dataRollRegridded.VSpec
		sZDR = dataRollRegridded.sZDR
		sSNR_H = dataRollRegridded.sSNR_H
		sSNR_V = dataRollRegridded.sSNR_V
	
	
	#- now move everything to 0 and name that rot (better for plotting)
	velMatrix = HSpec.values/HSpec.values*vel
	maxVelH = np.nanmax(velMatrix)#,axis=1)
	minVelH = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelH):
		maxVelH = 0
	if np.isnan(minVelH):
		minVelH = 0
	
	velMatrix = VSpec.values/VSpec.values*vel
	maxVelV = np.nanmax(velMatrix)#,axis=1)
	minVelV = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelV):
		maxVelV = 0
	if np.isnan(minVelV):
		minVelV = 0
		
	velMatrix = sZDR.values/sZDR.values*vel
	maxVelZDR = np.nanmax(velMatrix)#,axis=1)
	minVelZDR = np.nanmin(velMatrix)#,axis=1)
	if np.isnan(maxVelZDR):
		maxVelZDR = 0
	if np.isnan(minVelZDR):
		minVelZDR = 0
	#- now shift everything to 0, because we might get folding I have to increase Dopplervelocity range to have +-12
	HSpecRe = HSpec.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 
	vel2rot = int((-maxVelH/dv).round())
	if datasel.maxVelV > 0:
		HSpec_rot = HSpecRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		HSpec_rot = HSpecRe.roll(Vel=vel2rot,roll_coords=False)
	
	VSpecRe = VSpec.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 	
	vel2rot = int((-maxVelV/dv).round())
	if datasel.maxVelV > 0:
		VSpec_rot = VSpecRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		VSpec_rot = VSpecRe.roll(Vel=vel2rot,roll_coords=False)
		
	sZDRRe = sZDR.reindex({'Vel':newVel}, method = 'nearest', tolerance = 0.05) 
	vel2rot = int((-maxVelZDR/dv).round())
	if datasel.maxVelZDR > 0:
		sZDR_rot = sZDRRe.roll(Vel=vel2rot-1,roll_coords=False)
	else:
		sZDR_rot = sZDRRe.roll(Vel=vel2rot,roll_coords=False)
	return r,HSpec,VSpec,sZDR,sSNR_H,sSNR_V,maxVelH,minVelH,maxVelV,minVelV,maxVelZDR,minVelZDR,HSpec_rot,VSpec_rot,sZDR_rot



from pathlib import Path
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
from pdb import set_trace
from gfatpy.lidar.preprocessing import preprocess
from gfatpy.utils.plot import color_list


def binning(
    an: np.ndarray,
    pc: np.ndarray,
    pc_binning_range: tuple[float, float],
    pc_bin_width: float,
    channel_pc: str,
    datetime: str,
    plot_dir: Path,
    savefig: bool,
):
    def plot_binning(
        an: np.ndarray,
        pc: np.ndarray,
        median_pc: np.ndarray,
        mean_an: np.ndarray,
        std_an: np.ndarray,
        channel_pc: str,
        datetime: str,
        plot_dir: Path,
        savefig: bool,
    ):
        fig, ax = plt.subplots(figsize=[15, 5])
        ax.scatter(pc, an, c="grey", s=0.5, label="raw")
        ax.errorbar(
            median_pc,
            mean_an,
            yerr=std_an,
            linestyle="None",
            marker=".",
            color="red",
            label="binning",
        )
        ax.set_xlabel("PC signal, [MHz]")
        ax.set_ylabel("dc-corrected AN signal, [mV]")
        ax.set_title(f"{channel_pc} | {datetime}")
        ax.set_yscale("linear")
        ax.set_xlim(*pc_binning_range)
        ax.legend(fontsize=10, loc="upper right")
        if savefig:
            plt.savefig(plot_dir / f"binning_{channel_pc}_{datetime}.png")
        plt.close(fig)

    # Apply binning to pc and an
    bins = np.arange(
        pc_binning_range[0], pc_binning_range[1] + pc_bin_width, pc_bin_width
    )

    # Use numpy.digitize to get the bin indices for each element in signal_pc
    bin_indices = np.digitize(pc, bins)

    # Compute the median, mean, and standard deviation for each bin
    median_pc = np.array([np.median(pc[bin_indices == i]) for i in range(1, len(bins))])
    mean_an = np.array([np.mean(an[bin_indices == i]) for i in range(1, len(bins))])
    std_an = np.array([np.std(an[bin_indices == i]) for i in range(1, len(bins))])
    isnan = np.logical_and(np.isnan(median_pc), np.isnan(mean_an), np.isnan(std_an))

    # Plot binning
    plot_binning(
        an,
        pc,
        median_pc,
        mean_an,
        std_an,
        channel_pc,
        datetime,
        plot_dir,
        savefig=savefig,
    )

    return median_pc[~isnan], mean_an[~isnan], std_an[~isnan]


def get_valid_an_pc_values(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    pc_threshold: float,
    an_threshold: float,
) -> tuple[np.ndarray, np.ndarray]:
    gl_condition = np.logical_and(
        signal_an.values > an_threshold, signal_pc.values < pc_threshold
    )
    an = np.concatenate(signal_an.where(gl_condition).values, axis=0)
    pc = np.concatenate(signal_pc.where(gl_condition).values, axis=0)
    isnan = np.logical_and(np.isnan(an), np.isnan(pc))
    an = an[~isnan]
    pc = pc[~isnan]
    return an, pc


def cost_function(
    tau_range: np.ndarray,
    mean_an: np.ndarray,
    median_pc: np.ndarray,
    std_an: np.ndarray,
    channel_pc: str,
    datetime: str,
    savefig: bool,
    plot_dir: Path,
) -> tuple[np.ndarray, float]:
    
    def plot_cost_function(
        tau_range: np.ndarray,
        J: np.ndarray,
        channel_pc: str,
        plot_dir: Path,
        datetime: str,
        optimal_tau_index: int,
        savefig: bool,
    ):
        logJ = np.log(J)
        optimal_tau = tau_range[optimal_tau_index]
        fig, ax = plt.subplots(figsize=[15, 5])
        ax.plot(tau_range, logJ, c="blue", lw=2, label="J", zorder=1)
        ax.scatter(
            optimal_tau,
            logJ[optimal_tau_index],
            marker="p",
            s=50,
            color="yellow",
            label=f"optimal tau = {optimal_tau:.2f} ns",
            zorder=2,
        )
        ax.set_ylabel("log(J), [#]")
        ax.set_xlabel(r"$\tau$, [ns]")
        ax.minorticks_on()
        ax.grid(True, which="both", linestyle=":", linewidth=0.5, color=[0.3, 0.3, 0.3])
        ax.set_title(f"{channel_pc} | {datetime}")
        ax.legend(fontsize=10, loc="upper center")
        if savefig:
            plt.savefig(plot_dir / f"cost_function_{channel_pc}_{datetime}.png")
        plt.close(fig)

    J = np.nan * np.ones(len(tau_range))
    for idx, tau_ in enumerate(tau_range):
        dt_pc = median_pc / (1 - tau_ * median_pc * 1e-3)
        if not np.isnan(dt_pc).any() and not np.isnan(mean_an).any():
            try:
                fit_values = np.polyfit(dt_pc, mean_an, 1)
                virtual_an = np.polyval(fit_values, dt_pc)
            except Exception as e:                
                print(e)
                raise ValueError("Error in fitting binned an and pc")

            try:
                # Cost function
                idx_zero = std_an == 0
                J[idx] = np.sum(
                    (mean_an[~idx_zero] - virtual_an[~idx_zero]) ** 2
                    / std_an[~idx_zero] ** 2
                ) / len(mean_an)
            except Exception as e:                
                print(e)
                raise ValueError("Error in cost function")
    optimal_tau_index = np.log(J).argmin()

    plot_cost_function(
        tau_range=tau_range,
        J=J,
        channel_pc=channel_pc,
        plot_dir=plot_dir,
        datetime=datetime,
        optimal_tau_index=optimal_tau_index,
        savefig=savefig,
    )
    return J, tau_range[optimal_tau_index]


def plot_influence_of_dead_time_correction(
    tau_range: np.ndarray,
    mean_an: np.ndarray,
    median_pc: np.ndarray,
    channel_pc: str,
    datetime: str,
    plot_dir: Path,
    savefig: bool,
):
    fig, ax = plt.subplots(figsize=[15, 5])
    h0 = ax.scatter(mean_an, median_pc, label="original")
    colors = color_list(len(tau_range))
    handles = [None] * len(tau_range)
    for idx, tau_ in enumerate(tau_range):
        dt_pc = median_pc / (1 - tau_ * median_pc * 1e-3)
        handles[idx] = ax.scatter(
            mean_an, dt_pc, color=colors[idx], label=f"tau={tau_:.2f} ns"
        )
    ax.legend(
        [h0, handles[0], handles[-1]],
        ["original", f"tau={tau_range[0]:.2f} ns", f"tau={tau_range[-1]:.2f} ns"],
        fontsize=10,
        loc="upper left",
    )
    # ax.legend([h0, handles[0], handles[-1]], fontsize=10, loc="upper left")
    ax.set_ylabel("dt-corrected PC signal, [MHz]")
    ax.set_xlabel("dc-corrected AN signal, [mV]")
    ax.set_title(datetime)
    ax.grid(True, which="both", linestyle=":", linewidth=0.5, color=[0.3, 0.3, 0.3])
    if savefig:
        plt.savefig(plot_dir / f"pc_taus_{channel_pc}_{datetime}.png")
    plt.close(fig)


def plot_gluing_thresholds(
    signal_an: xr.DataArray,
    signal_pc: xr.DataArray,
    pc_threshold: float,
    an_threshold: float,
    datetime: str,
    plot_dir: Path,
    savefig: bool,
    range_limits: tuple[float, float],
):
    channel_pc = signal_pc.name.split("_")[-1]

    fig, ax = plt.subplots(figsize=[15, 5])
    lan = signal_an.plot.line(x="range", c="grey", lw=0.5, label="an")
    lpc = signal_pc.plot.line(x="range", c="black", lw=0.5, label="pc")

    l1 = plt.hlines(
        pc_threshold, xmin=0, xmax=10000, color="darkred", ls="--", label="Cmax"
    )
    l2 = plt.hlines(
        an_threshold,
        xmin=0,
        xmax=10000,
        color="violet",
        ls="--",
        label="bg_threshold_an",
    )

    ax.set_title(f"{channel_pc} | {datetime}")
    ax.set_yscale("log")
    ax.set_ylim(0.01, 3000)
    ax.set_xlim(*range_limits)
    ax.legend(handles=[lan[0], lpc[0], l1, l2], fontsize=10, loc="upper right")

    if savefig:
        plt.savefig(plot_dir / f"gluing_thresholds_{channel_pc}_{datetime}.png")
    plt.close(fig)


def dead_time_finder_by_channel(
    lidar: Path,
    channel_an: str,
    channel_pc: str,
    pc_threshold: float,
    an_threshold: float,
    tau_range: list,
    pc_binning_range: tuple[float, float],
    pc_bin_width: float,
    plot_dir: Path,
    savefig: bool,
    range_limits: tuple[float, float],
) -> float:
    datetime = lidar.time.values[0].astype(str).split("T")[0].replace("-", "")
    signal_an = lidar[f"signal_{channel_an}"]
    signal_pc = lidar[f"signal_{channel_pc}"]

    plot_gluing_thresholds(
        signal_an=signal_an,
        signal_pc=signal_pc,
        pc_threshold=pc_threshold,
        an_threshold=an_threshold,
        datetime=datetime,
        plot_dir=plot_dir,
        savefig=savefig,
        range_limits=range_limits,
    )

    # get valid an and pc values
    an, pc = get_valid_an_pc_values(
        signal_an=signal_an,
        signal_pc=signal_pc,
        pc_threshold=pc_threshold,
        an_threshold=an_threshold,
    )

    if len(an) < 100 or len(pc) < 100:
        print(f"Empty arrays for {channel_pc} | {datetime}")
        return np.nan

    # apply binning
    median_pc, mean_an, std_an = binning(
        an=an,
        pc=pc,
        pc_binning_range=pc_binning_range,
        pc_bin_width=pc_bin_width,
        channel_pc=channel_pc,
        datetime=datetime,
        plot_dir=plot_dir,
        savefig=savefig,
    )
    
    # Plot influence of dead time correction
    plot_influence_of_dead_time_correction(
        tau_range=tau_range,
        mean_an=mean_an,
        median_pc=median_pc,
        channel_pc=channel_pc,
        datetime=datetime,
        plot_dir=plot_dir,
        savefig=savefig,
    )

    # Retrieve cost function
    J, optimal_tau = cost_function(
        tau_range=tau_range,
        mean_an=mean_an,
        median_pc=median_pc,
        std_an=std_an,
        channel_pc=channel_pc,
        datetime=datetime,
        savefig=savefig,
        plot_dir=plot_dir,
    )    

    #Check if optimal tau is out of range
    if optimal_tau == tau_range[0] or optimal_tau == tau_range[-1]:
        print(f'Warning! {datetime} | {channel_pc} : tau out of range')
        return np.nan
    print(f"{datetime} | Optimal tau for {channel_pc} = {optimal_tau} ns")
    
    return optimal_tau


def estimate_daily_dead_time(
    file: Path | str,
    tau_dir: Path | str | None = None,
    tau_range: list | np.ndarray = np.arange(2, 10, 0.1),
    crop_range: tuple = (0, 15000),
    pc_threshold: float = 50,
    an_threshold: float = 0.1,
    pc_binning_range: tuple = (0, 60),
    pc_bin_width=1,
    range_limits: tuple[float, float] = (0, 10000),
    savefig: bool = False,
) -> Path:
    # Management file type
    if isinstance(file, str):
        file = Path(file)
    if file.exists() == False:
        raise FileNotFoundError(f"{file} does not exist")
    if file.suffix != ".nc":
        raise TypeError(f"{file} is not a netcdf file")

    # Management tau_dir type
    if isinstance(tau_dir, str):
        tau_dir = Path(tau_dir)
    if tau_dir is None:
        tau_dir = Path.cwd()

    # Management tau_range type
    if isinstance(tau_range, list):
        tau_range = np.array(tau_range)
    if tau_range.ndim != 1:
        raise ValueError(f"tau_range must be a 1D array")

    # Read
    lidar = preprocess(
        file,
        apply_dt=False,
        save_bg=False,
        save_dc=True,
        apply_bz=False,
        crop_ranges=crop_range,
    )
    datetime = lidar.time.values[1].astype(str).split("T")[0].replace("-", "")
    
    if savefig:
        plot_dir = tau_dir / f"plots_{datetime}"
        if plot_dir.exists() == False:
            try:
                plot_dir.mkdir()
            except Exception as e:
                print(e)
                raise ValueError(f"Error creating {plot_dir}")

    # Create tau dictionary
    optimal_taus = {}

    for channel_ in [
        channel_ for channel_ in lidar.channel.values if channel_.endswith("a")
    ]:
        channel_an, channel_pc = channel_, channel_.replace("a", "p")
        if channel_pc in lidar.channel:

            optimal_taus[channel_pc] = dead_time_finder_by_channel(
                lidar=lidar,
                channel_an=channel_an,
                channel_pc=channel_pc,
                pc_threshold=pc_threshold,
                an_threshold=an_threshold,
                pc_binning_range=pc_binning_range,
                pc_bin_width=pc_bin_width,
                plot_dir=plot_dir,
                savefig=savefig,
                tau_range=tau_range,
                range_limits=range_limits,
            )

    # Save optimal_taus
    tau_da = xr.DataArray(
        list(optimal_taus.values()),
        coords=[list(optimal_taus.keys())],
        dims=["channel"],
    )
    tau_da.attrs["date"] = datetime
    output_file = tau_dir / f"alh_dead-time_{datetime}.nc"
    tau_da.to_netcdf(output_file)
    return output_file



from pathlib import Path
import numpy as np

from rpgpy import read_rpg
from rpgpy.spcutil import scale_spectra

class rpg():
    def __init__(self, path: Path):
        self.path = path        
        self.type = path.name.split(".")[0].split("_")[-1]
        self.level = int(path.name.split(".")[-1][-1])
        self._raw = None
        self._header = None
        self._data = None

    @property
    def raw(self) -> dict:
        if self._raw is None:            
            _, self._raw = read_rpg(self.path)
        return self._raw
    
    @property
    def header(self) -> dict:
        if self._header is None:            
            self._header, _ = read_rpg(self.path)
        return self._header

    @property
    def data(self) -> dict:
        if self._data is None:            
            self._data = self.enhanced_data(self.header, self.raw)
        return self._data

    @classmethod
    def enhanced_data(cls, header: dict, raw: dict) -> dict:
        data = raw.copy()        
        spec_tot = scale_spectra(raw["TotSpec"], header["SWVersion"])
        data['VSpec'] = spec_tot - raw["HSpec"] - 2 * raw["ReVHSpec"]
        data['sZDR'] = 10*np.log10(data['HSpec']) - 10*np.log10(data['VSpec'])        
        data['sZDRmax'] = np.max(data['sZDR'], axis=2)
        return data
    
    # def merge_chirps(self) -> dict:
    #     if self.level == 0:
    #         data = merge_chirps_lv0(self.header, self.raw)
    #     return data


    # def plot(self, variable: str | None = None, **kwargs):        
    #     if self.type == "ZEN":
    #         self.plot_zen(variable, **kwargs)
    #     elif self.type == "PPI":
    #         self.plot_ppi(variable, **kwargs)
    #     elif self.type == "RHI":
    #         self.plot_rhi(variable, **kwargs)
    #     else:
    #         raise ValueError(f"Type {self.type} is not valid")

    # def plot_zen(self, variable: str | list[str] | None = None, **kwargs):
        
    #     range_limits = kwargs.get('range_limits', (0, 12.))
        
        
    #     if variable is None:
    #         variables_to_plot = ["dBZe"]
    #     if isinstance(variable, str):
    #         variables_to_plot = [variable]
    #     elif isinstance(variable, list):
    #         variables_to_plot = variable
    #     else:
    #         raise ValueError(f"Variable {variable} is not valid")        
        
    #     data = self.data
    #     data['range'] = data['range']/1e3
    #     list_of_figs, list_of_paths = [], []        
    #     for var in variables_to_plot:
    #         if var not in self.data:
    #             raise ValueError(f"Variable {var} is not in the file")
            
    #         vmin, vmax = RADAR_PLOT_INFO['limits'][var]
    #         fig, ax = plt.subplots(
    #             figsize=(10, 7)
    #         )  # subplot_kw=dict(projection='polar')

    #         pcm = self.data[f"{var}"].plot(x='time', ax=ax, vmin=vmin, vmax=vmax)
            
    #         ax.set_xlabel("Time, [UTC]")
    #         ax.set_ylim(*range_limits)
    #         ax.set_ylabel("Range, [km]")

    #         #Get current colorbar and shrink it 0.7, and set 'units' as label
            

    #         # cbar = plt.colorbar(pcm, ax=ax, shrink=0.7)
    #         if "units" in data[var].attrs:
    #             units = data[var].attrs['units']
    #         else:
    #             units = "?"
    #         pcm.set_label(f"{var}, [{units}]")
            
    #         ax.set_title(f"{data.time.values[0]}")            
            
    #         fig.tight_layout()
    #         list_of_figs.append(fig)
    #         if kwargs.get('savefig', False):
    #             output_dir = kwargs.get("output_dir", Path.cwd())   
    #             filepath = output_dir / self.path.name.replace(".nc", f"_{var}.png")
    #             dpi = kwargs.get("dpi", 300)    
    #             fig.savefig(filepath, dpi=dpi)
    #             plt.close(fig)
    #             list_of_paths.append(filepath)
    #     return list_of_figs, list_of_paths


    # def plot_ppi(self, variable: str | None = None, **kwargs) -> tuple[list[Figure], list[Path]]:
    #     if variable is None:
    #         variables_to_plot = ["dBZe"]
    #     elif isinstance(variable, str):
    #         variables_to_plot = [variable]
    #     elif isinstance(variable, list):
    #         variables_to_plot = variable
    #     else:
    #         raise ValueError(f"Variable {variable} is not valid")

    #     # sort data['azimuth'] as increasing and sort the rest of the data accordingly
    #     data = self.data.sortby("azimuth")
    #     x, y = ppi_to_cartessian(data["range"], data["azimuth"], data["elevation"])
    #     mdata = data.mean(dim="time")
    #     list_of_figs, list_of_paths = [], []
    #     for var in variables_to_plot:
    #         if var not in mdata:
    #             raise ValueError(f"Variable {var} is not in the file")
    #         vmin, vmax = RADAR_PLOT_INFO["limits"][var]
    #         fig, ax = plt.subplots(
    #             figsize=(10, 10)
    #         )  # subplot_kw=dict(projection='polar')
    #         pcm = ax.pcolormesh(
    #             x/1e3, y/1e3, data[var].values.T, vmin=vmin, vmax=vmax, shading="gouraud"
    #         )
    #         #Include title with time and elevation angle
    #         ax.set_xlim(*RADAR_PLOT_INFO["limits"]["ppi"]["x"])
    #         ax.set_ylim(*RADAR_PLOT_INFO["limits"]["ppi"]["y"])
    #         ax.set_xlabel("East-West distance from radar [km]")
    #         ax.set_ylabel("North-South distance from radar [km]")
    #         circular_grid(ax, radius = ax.get_xticks())
    #         ax.set_aspect("equal")
    #         cbar = plt.colorbar(pcm, ax=ax, shrink=0.7)
    #         if "units" in data[var].attrs:
    #             units = data[var].attrs['units']
    #         else:
    #             units = "?"
    #         cbar.set_label(f"{var}, [{units}]")
    #         ax.set_title(f"{data.time.values[0]}. Elevation: {data.elevation.values[0]}")            
    #         fig.tight_layout()
    #         list_of_figs.append(fig)
    #         if kwargs.get('savefig', False):
    #             output_dir = kwargs.get("output_dir", Path.cwd())
    #             filepath = output_dir / self.path.name.replace(".nc", f"_{var}.png")
    #             dpi = kwargs.get("dpi", 300)    
    #             fig.savefig(filepath, dpi=dpi)
    #             plt.close(fig)
    #             list_of_paths.append(filepath)
    #     return list_of_figs, list_of_paths


    # def plot_rhi(self, variable: str | None = None, **kwargs):
    #     if variable is None:
    #         variables_to_plot = ["dBZe"]
    #     elif isinstance(variable, str):
    #         variables_to_plot = [variable]
    #     elif isinstance(variable, list):
    #         variables_to_plot = variable
    #     else:
    #         raise ValueError(f"Variable {variable} is not valid")

    #     label_angle = kwargs.get("label_angle", None)
    #     circular_grid = kwargs.get("circular_grid", False)

    #     # sort data['azimuth'] as increasing and sort the rest of the data accordingly
    #     data = self.data.sortby("elevation")
    #     constante_azimuth_angle = data["azimuth"].values[0]
    #     x, y = rhi_to_cartessian(data["range"], data["azimuth"], data["elevation"])
    #     mdata = data.mean(dim="time")
    #     list_of_figs, list_of_paths = [], []
    #     for var in variables_to_plot:
    #         if var not in mdata:
    #             raise ValueError(f"Variable {var} is not in the file")
    #         vmin, vmax = RADAR_PLOT_INFO["limits"][var]
    #         fig, ax = plt.subplots(
    #             figsize=(10, 10)
    #         )  # subplot_kw=dict(projection='polar')
    #         pcm = ax.pcolormesh(x/1e3, y/1e3, 
    #             data[var].values.T, 
    #             vmax = vmin,
    #             vmin = vmax,
    #             shading='gouraud'
    #         )
    #         ax.set_title(f"{str(data.time.values[0]).split('.')[0]} | Azimuth: {constante_azimuth_angle:.1f}°")
    #         ax.set_xlim(*RADAR_PLOT_INFO["limits"]["rhi"]["x"])
    #         ax.set_ylim(*RADAR_PLOT_INFO["limits"]["rhi"]["y"])
    #         ax.set_xlabel(f"Distance from radar, [km]")
    #         ax.set_ylabel("Height distance from radar, [km]")
    #         if circular_grid:
    #             if label_angle is None:
    #                 label_angle = np.min(self.data["elevation"].values)-5
    #             circular_grid(ax, radius = ax.get_xticks(), label_angle=label_angle)
    #         ax.set_aspect("equal")
    #         cbar = plt.colorbar(pcm, ax=ax, shrink=0.7)
    #         if "units" in data[var].attrs:
    #             units = data[var].attrs['units']
    #         else:
    #             units = "?"
    #         cbar.set_label(f"{var}, [{units}]")
    #         fig.tight_layout()
    #         list_of_figs.append(fig)
    #         if kwargs.get('savefig', False):
    #             output_dir = kwargs.get("output_dir", Path.cwd())
    #             filepath = output_dir / self.path.name.replace(".nc", f"_{var}.png")
    #             dpi = kwargs.get("dpi", 300)    
    #             fig.savefig(filepath, dpi=dpi)
    #             plt.close(fig)
    #             list_of_paths.append(filepath)
    #     return list_of_figs, list_of_paths



    # def read_rpg_spectra(self):
    #     pass

    # def read_rpg_spectra_header(self):
    #     pass

    # def __str__(self) -> str:
    #     return super().__str__()
    


from pathlib import Path
from gfatpy.utils.io import read_yaml

SCC_INFO = read_yaml(Path(__file__).parent.absolute() / "info.yml")
MAX_EXEC_TIME = int(1200)  # 3600

__all__ = ["SCC_INFO", "MAX_EXEC_TIME"]



from matplotlib import pyplot as plt
import pandas as pd
import xarray as xr
import numpy as np
from scipy.integrate import cumtrapz


from gfatpy.lidar.retrieval.klett import klett_rcs
from gfatpy.lidar.retrieval.raman import retrieve_backscatter, retrieve_extinction
from gfatpy.lidar.utils.utils import refill_overlap, signal_to_rcs
from gfatpy.atmo.rayleigh import molecular_properties


def overlap_iterative(elastic_signal: xr.DataArray, 
                      raman_signal: xr.DataArray,                      
                      meteo_profiles: pd.DataFrame, 
                      particle_lidar_ratio: float, 
                      reference: tuple[float, float] = (8000,8500), 
                      wavelengths: tuple[float, float] = (532, 531), 
                      beta_aer_ref: float = 0, 
                      min_overlap_range: float = 200, #minimum range of overlap function
                      particle_angstrom_exponent: float = 0.,
                      iteration_limit: int= 10,
                      **kwargs: dict
                      ) -> tuple[xr.DataArray, np.ndarray, np.ndarray]:
    """Retrieve overlap function by means of the Wandinger's method. 

    Args:
        elastic_signal (xr.DataArray): Elastic lidar signal at wavelength `wavelengths[0]`.
        raman_signal (xr.DataArray): Raman lidar signal at wavelength `wavelengths[1]`.        
        meteo_profiles (pd.DataFrame): from  `gfatpy.atmo.atmo.generate_meteo_profiles` with pressure and temperature data.
        particle_lidar_ratio (float): Particle lidar ratio to be used in the Klett inversion. It shall be chosen so Klett backscatter profile fits Raman backscatter profile once full overlap is reached.    
        reference (tuple[float, float], optional): Reference range required by the Klett and Raman inversions. Defaults to (8000,8500).
        wavelengths (tuple[float, float], optional): Elastic and Raman wavelengths. Defaults to (532, 531).
        beta_aer_ref (float, optional): Particle backscatter coefficient at `reference range` in m^-1sr^-1. Defaults to 0 m^-1sr^-1.
        particle_angstrom_exponent (float, optional): Particle extinction-related Angstrom exponent. It is used in the Raman inversion. Defaults to 0.
        min_overlap_range (float, optional): Overlap function will be forced to zero below [m]. Defaults to 200 m.

    Returns:
        xr.DataArray: overlap function.
    """    
    #Check z in elastic and raman signals (the same for both signals)
    if len(elastic_signal.range) != len(raman_signal.range):
        raise   ValueError("elastic and Raman signals must have the same 'range' dimension.")
        
    if len(elastic_signal.range) != len(meteo_profiles['height']):
        raise   ValueError("Signals range dimension and meteo_profiles['heights'] must have the same size.")

    z = elastic_signal.range.values
    
    mol_properties = molecular_properties(wavelengths[0], meteo_profiles['pressure'], meteo_profiles['temperature'], heights=z)
    molecular_backscatter = mol_properties['molecular_beta'].values
    pressure = meteo_profiles['pressure'].values
    temperature = meteo_profiles['temperature'].values

    #Beta Klett
    elastic_backscater = klett_rcs(signal_to_rcs(elastic_signal, z),
                        z,
                        molecular_backscatter,
                        lr_part=particle_lidar_ratio,
                        reference = reference
                        ) 
    
    extinction = retrieve_extinction(
        raman_signal.values, 
        z, 
        wavelengths, 
        pressure, 
        temperature, 
        reference=reference,
    ) 

    #Refill extinction
    extinction = refill_overlap(extinction, z, kwargs.get('fulloverlap_height', 600.))

    fig, ax = plt.subplots()
    ax.plot(extinction*1e6, z)
    ax.set_xlabel('Extinction [m^-1]')
    ax.set_ylabel('Altitude [m]')
    ax.set_xlim(0, 500)
    fig.savefig('extinction.png')

    raman_backscatter = retrieve_backscatter(
        raman_signal.values, 
        elastic_signal.values,
        extinction, 
        z, 
        wavelengths, 
        pressure, 
        temperature, 
        particle_angstrom_exponent=particle_angstrom_exponent) #Fine mode has been chosen in Amstrong exponent 
    
    i = 0
    corrected_signal = np.zeros([iteration_limit, len(elastic_signal)])
    delta_overlap = np.zeros(corrected_signal.shape)
    elastic_backscatter_matrix = np.zeros(corrected_signal.shape)
    corrected_signal[i,:] = elastic_signal
    elastic_backscatter_matrix[i,:] = elastic_backscater
    delta_overlap[i,:] = (raman_backscatter - elastic_backscatter_matrix[i,:]) / (raman_backscatter + molecular_backscatter)

    #Iteraciones restantes
    for i in range(1, iteration_limit):
        corrected_signal[i,:] = corrected_signal[i-1, :] * (1 + delta_overlap[i-1,:])
        rcs_ = signal_to_rcs(corrected_signal[i,:], z)
        elastic_backscatter_matrix[i,:] = klett_rcs(rcs_,
                                z,
                                molecular_backscatter,
                                lr_part=particle_lidar_ratio,
                                beta_aer_ref=beta_aer_ref
                                )
        
        delta_overlap[i,:] = (raman_backscatter - elastic_backscatter_matrix[i,:]) / (raman_backscatter + molecular_backscatter)
    overlap_function = (corrected_signal[0,:] / corrected_signal[-1,:]) 

    #Set to zero the values of overlap function below 200 m
    overlap_function[:np.where(z > min_overlap_range)[0][0]] = 0
    
    #Search first 1 in overlap function
    first_1 = np.where(overlap_function >= 1)[0][0]

    #Set constant profile to 1 form first 1 up
    overlap_function[first_1:] = 1

    return overlap_function, elastic_backscatter_matrix, raman_backscatter

def overlap_function_explicit(elastic_rcs: xr.DataArray, 
                              raman_rcs: xr.DataArray,
                              wavelengths: tuple[float, float],
                              reference_height: float,
                              particle_lidar_ratio: float,
                              meteo_profiles: pd.DataFrame, 
                              **kwargs: dict) -> np.ndarray:
    """Retrieve overlap function by means of the explicit method.

    Args:

        - elastic_rcs (xr.DataArray): Elastic range corrected signal at wavelength `wavelengths[0]`.
        - raman_rcs (xr.DataArray): Raman range corrected signal at wavelength `wavelengths[1]`.
        - wavelengths (tuple[float, float]): Elastic and Raman wavelengths.
        - reference_height (float): Aerosol-free region reference height in meters.
        - particle_lidar_ratio (float): Assumed particle lidar ratio.
        - meteo_profiles (pd.DataFrame): Meteo profiles with pressure and temperature data from `gfatpy.atmo.atmo.generate_meteo_profiles`.

    Returns:

        - np.ndarray: overlap function.

    References:

        - Comerón et al., 2023: https://doi.org/10.5194/amt-16-3015-2023
    """    
    
    r = elastic_rcs.range.values
    dr = r[1] - r[0]

    elastic_mol_properties = molecular_properties(wavelengths[0], meteo_profiles['pressure'], meteo_profiles['temperature'], heights=z)
    raman_mol_properties = molecular_properties(wavelengths[1], meteo_profiles['pressure'], meteo_profiles['temperature'], heights=z)

    
    # Computation of beta molecular for elastic and Raman signal
    beta_mol_el = elastic_mol_properties['molecular_beta']*1000 # it has to be in km-1*sr-1
    beta_mol_ra = raman_mol_properties['molecular_beta']*1000

    # Cálculo inverso de beta molecular, a partir de la señal Rayleigh
    X = elastic_rcs.values # elastic signal
    X_Ray = X_Ray # molecular backscatter normalized to RCS

    XR = raman_rcs.values # raman signal
    XR_Ray = XR_Ray # molecular backscatter normalized to RCS


    ref_height = reference_height / 1e3 # Reference height (km)
    idx = np.round(ref_height / dr).astype(int)
    X_ref = X_Ray[idx]
    XR_ref = XR_Ray[idx]
    beta_mol_ref = beta_mol_el[idx]

    nXR = XR/XR_ref
    nbeta_mol = beta_mol_el/beta_mol_ref

    Sa = particle_lidar_ratio # Assumed Lidar Ratio
    Sm_el = elastic_mol_properties['molecular_lidar_ratio']
    # Sm_ra = raman_mol_properties['molecular_lidar_ratio']

    beta_mol_el = beta_mol_el[:idx]
    beta_mol_ra = beta_mol_ra[:idx]
    nbeta_mol = nbeta_mol[:idx]
    nXR = nXR[:idx]
    X = X[:idx]

    fexp = np.exp(-2 * (Sa - Sm_el) * np.flip(cumtrapz(np.flip(beta_mol_el), initial=0, dx=dr)))

    #Not used because its efect is negligible
    # fexp_dif_ext_mol = np.exp(- Sm_el * np.flip(cumtrapz(np.flip(beta_mol_el - beta_mol_ra), initial=0, dx=dr)))

    g = (nbeta_mol/nXR) * fexp

    phi = (2 * beta_mol_el / (nXR * X_ref)) * fexp

    psi = (X * Sa) / fexp

    phi_psi = phi * psi

    fexp_phi_psi = np.exp(np.flip(cumtrapz(np.flip(phi_psi), initial=0, dx=dr)))

    fint_g_psi = np.flip(cumtrapz(np.flip(g * psi /fexp_phi_psi), initial=0, dx=dr))

    f = g + phi * fexp_phi_psi * fint_g_psi

    Oe = 1/f

    return Oe


from datetime import datetime
from pathlib import Path
import xarray as xr 

from matplotlib import pyplot as plt
from matplotlib.figure import Figure
import numpy as np

from gfatpy.radar.rpg_nc import rpg
from gfatpy.utils.utils import parse_datetime


class nebula():
    def __init__(self, ka_nc_path: Path, ww_nc_path: Path):
        self.ka = rpg(ka_nc_path)
        self.ww = rpg(ww_nc_path)        
        self._level = None
        self.nb = self.ka
        self.nb._data = self.combined_data()

    @property
    def level(self) -> int:
        if self.ka.level != self.ww.level:
            raise ValueError("Ka- and W-band measurement levels do not match.")        
        self._level = self.ka.level
        return self._level

    def combined_data(self) -> xr.Dataset:
        if self.level == 1:
            if self.nb._data is None:
                self.nb._data["DWR"] = 10 * np.log10(self.ka.data["Ze"] / self.ww.data["Ze"])
                self.nb._data["DWR"].attrs["units"] = "dB"
                self.nb._data["DWR"].attrs["long_name"] = "Ka-W DWR"

                self.nb._data["DDV"] = self.ka.data["v"] - self.ww.data["v"]
                self.nb._data["DDV"].attrs["units"] = "m/s"
                self.nb._data["DDV"].attrs["long_name"] = "Ka-W DDV"
        return self.nb._data

    def quicklook(self, variable: str | list[str] | None = None, **kwargs):
        self.ka.quicklook(variable, **kwargs)
        self.ww.quicklook(variable, **kwargs)

    def plot_profile(
        self,
        target_time: datetime | np.datetime64,
        range_limits: float,
        variable: str | list[str] | None = None,
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        _kwargs = kwargs.copy()        
        _kwargs["savefig"] = False
        _kwargs["color_list"] = [kwargs.get("ka_color", "red")]

        fig1, _ = self.ka.plot_profile(
            target_times=target_time,
            range_limits=range_limits,
            variable=variable,
            **_kwargs,
        )

        _kwargs["fig"] = fig1
        _kwargs["color_list"] = [kwargs.get("w_color", "blue")]
        fig, _ = self.ww.plot_profile(
            target_times=target_time,
            range_limits=range_limits,
            variable=variable,
            **_kwargs,
        )

        ka_line, w_line = fig.findobj(plt.Line2D)[0], fig.findobj(plt.Line2D)[1]
        ka_line.set_label(f"Ka-band")
        w_line.set_label(f"W-band")

        ax = fig.get_axes()[0]
        ax.legend()

        if kwargs.get("savefig", False):
            filepath = (
                kwargs.get("output_dir", Path.cwd())
                / f"profile_{variable}_{target_time:%Y%m%dT%H%M%S}.png"
            )
            fig.savefig(filepath, dpi=kwargs.get("dpi", 300))
        return fig, filepath

    def plot_timeseries(
        self,
        target_range: float,
        time_limits: tuple[datetime | np.datetime64, datetime | np.datetime64] | None = None,
        variable: str | list[str] | None = None,
        **kwargs,
    ) -> tuple[Figure, Path | None]:
        
        _kwargs = kwargs.copy()        
        _kwargs["savefig"] = False
        _kwargs["color_list"] = [kwargs.get("ka_color", "red")]

        if  time_limits is None:
            time_limits = (parse_datetime(self.ka.data.time[0].values), parse_datetime(self.ka.data.time[-1].values))

        fig1, _ = self.ka.plot_timeseries(
            target_ranges=target_range,
            time_limits=time_limits,
            variable=variable,
            **_kwargs,
        )

        _kwargs["fig"] = fig1
        _kwargs["color_list"] = [kwargs.get("w_color", "blue")]
        fig, _ = self.ww.plot_timeseries(
            target_ranges=target_range,
            time_limits=time_limits,
            variable=variable,
            **_kwargs,
        )

        ka_line, w_line = fig.findobj(plt.Line2D)[0], fig.findobj(plt.Line2D)[1]
        ka_line.set_label(f"Ka-band")
        w_line.set_label(f"W-band")

        ax = fig.get_axes()[0]
        variable_string = self.ka.data[variable].attrs['long_name'].replace('W-band','').replace("Ka-band",'')
        ax.set_ylabel(f"{variable_string}, [{self.ka.data[variable].attrs['units']}]")
        ax.legend()

        if kwargs.get("savefig", False):
            filepath = (
                kwargs.get("output_dir", Path.cwd())
                / f"timeseries_{variable}_{time_limits[0]:%Y%m%dT%H%M%S}_{time_limits[-1]:%Y%m%dT%H%M%S}_{target_range}.png"
            )
            fig.savefig(filepath, dpi=kwargs.get("dpi", 300))
        return fig, filepath


    def plot_spectrum(
        self, target_time: datetime | np.datetime64, target_range: float, **kwargs
    ) -> tuple[Figure, Path | None]:
        _kwargs = kwargs.copy()
        _kwargs.pop("output_dir")
        _kwargs.pop("savefig")
        _kwargs["color"] = kwargs.get("ka_color", "red")
        fig, _ = self.ka.plot_spectrum(target_time, target_range, **_kwargs)

        _kwargs["fig"] = fig
        _kwargs["color"] = kwargs.get("w_color", "blue")
        fig, _ = self.ww.plot_spectrum(target_time, target_range, **_kwargs)
        ka_line, w_line = fig.findobj(plt.Line2D)[0], fig.findobj(plt.Line2D)[2]
        ka_line.set_label(f"Ka-band {ka_line.get_label()}")
        w_line.set_label(f"W-band {w_line.get_label()}")

        ax = fig.get_axes()[0]
        ax.legend()

        if kwargs.get("savefig", False):
            filepath = (
                kwargs.get("output_dir", Path.cwd())
                / f"{target_time:%Y%m%dT%H%M%S}_{target_range:.0f}_spectrum.png"
            )
            fig.savefig(filepath, dpi=kwargs.get("dpi", 300))

        return fig, filepath
